----------------------------------------------------------------------------
-- @Author: ViperGTS96-----------------------------------------------------
----------------------------------------------------------------------------
--------------------"The simplest design is the best design." --------------
----------------------------------------------------------------------------
--This design of handling wheel particle systems is reserved by ViperGTS96--
----------------------------------------------------------------------------

waterBodyFX = {};
waterBodyFX.dir = ""
local modDescFile = loadXMLFile("modDesc", g_currentModDirectory.. "modDesc.xml");
waterBodyFX.title = getXMLString(modDescFile, "modDesc.title.en");
waterBodyFX.author = getXMLString(modDescFile, "modDesc.author");
waterBodyFX.version = getXMLString(modDescFile, "modDesc.version");
waterBodyFX.printVehiclesToLog = false; --(debugging)
waterBodyFX.maxWheelLimit = 10;
waterBodyFX.psSpeedRatio = 25000; --Divedend
delete(modDescFile);

function waterBodyFX:setDir(dir) waterBodyFX.dir = dir; end;

function waterBodyFX.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Washable, specializations)
		and SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function waterBodyFX.registerEventListeners(vehicleType)
	for _, functionName in pairs( { "onPostLoad", "onUpdate", "onLeave" } ) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, waterBodyFX);
	end;
end;

function waterBodyFX:onPostLoad(savegame)
	if waterBodyFX.dir == "" then return; end;
	if self.spec_wheels == nil then return; end;
	local wheels = self.spec_wheels.wheels;
	local crawlers = self.spec_crawlers;
	if wheels ~= nil and #wheels > 0 then
		self.waterBody_FX = {};
		self.waterBody_FX.tooManyWheelsForwater = false;
		if self.spec_motorized ~= nil then
			self.waterBody_FX.defaultAccelLimit =self:getMotor():getAccelerationLimit();
		end;
		if crawlers ~= nil and #crawlers.crawlers > 0 then
			crawlers = crawlers.crawlers;
			self.waterBody_FX.crawlers = #crawlers;
		end;
		if #wheels > waterBodyFX.maxWheelLimit then
			self.waterBody_FX.tooManyWheelsForwater = true;
			return;
		else
			for index, wheel in pairs(wheels) do
				local wheelLinkNode = wheel.driveNode;
				if wheelLinkNode == nil then wheelLinkNode = wheel.repr; end;
				wheel.waterBodyFX_isCrawler = waterBodyFX:getIsCrawler(self, wheel, crawlers);
				local hasIssue, fix = waterBodyFX:getHasIssue(self, index, wheel);
				if not hasIssue then
					local water_psData = {};
					local scaleFactor = 1.0;
					wheel.water_psF = {};
					wheel.water_psR = {};
					wheel.waterBodyFXisActive = false;
					wheel.waterBodyFXMainNode = createTransformGroup("waterFXNode");
					link(wheel.node, wheel.waterBodyFXMainNode);
					local rDX1,rDY1,rDZ1 = localToWorld(self.components[1].node, 0,0,1);
					local rDX2,rDY2,rDZ2 = localToWorld(self.components[1].node, 0,1,0);
					setWorldTranslation(wheel.waterBodyFXMainNode, unpack({localToWorld(wheelLinkNode,0,-wheel.radius,0)}));
					setDirection(wheel.waterBodyFXMainNode,rDX1,rDY1,rDZ1,rDX2,rDY2,rDZ2);
					if fix == 1 then local bx,by,bz = getTranslation(wheels[index-1].waterBodyFXMainNode);
					setTranslation(wheel.waterBodyFXMainNode, -bx,by,bz);	end;
					water_psData.psFile = "particleSystems/wheelWaterPS.i3d";
					water_psData.posX = 0;
					water_psData.posY = 0;
					water_psData.posZ = -wheel.radius;
					water_psData.worldSpace = false;
					ParticleUtil.loadParticleSystemFromData(water_psData, wheel.water_psF, nil, false, nil, waterBodyFX.dir, wheel.waterBodyFXMainNode);
					water_psData.posZ = wheel.radius;
					if wheel.waterBodyFX_isCrawler then water_psData.posZ = crawlers[1].trackWidth*3 end;
					ParticleUtil.loadParticleSystemFromData(water_psData, wheel.water_psR, nil, false, nil, waterBodyFX.dir, wheel.waterBodyFXMainNode);
					local rX,rY,rZ = getRotation(wheel.water_psR.shape);
					setRotation(wheel.water_psR.shape, rX,rY-math.rad(180),rZ);
					setScale(wheel.water_psF.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					setScale(wheel.water_psR.shape, wheel.width * scaleFactor, wheel.radius * scaleFactor, wheel.radius * scaleFactor);
					wheel.water_psF.resetPos = {getTranslation(wheel.water_psF.shape)}
					wheel.water_psR.resetPos = {getTranslation(wheel.water_psR.shape)}
					local ir = 1;
					local il = 1;
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							additionalWheel.waterBodyFX_isCrawler = false;
							if additionalWheel.tireType == WheelsUtil.getTireType("crawler") then
								additionalWheel.waterBodyFX_isCrawler = true;
							end;
							local offset = 0.11*wheel.radius;
							if wheel.xOffset ~= nil and wheel.xOffset ~= 0 then offset = wheel.xOffset; end;
							if additionalWheel.isLeft then
								water_psData.posX = ((wheel.width*il) + offset);
								il = il + 1;
							else
								water_psData.posX = -((wheel.width*ir) + offset);
								ir = ir + 1;
							end;
							additionalWheel.water_psF = {};
							additionalWheel.water_psR = {};
							water_psData.posZ = -wheel.radius;
							ParticleUtil.loadParticleSystemFromData(water_psData, additionalWheel.water_psF, nil, false, nil, waterBodyFX.dir, wheel.waterBodyFXMainNode);
							water_psData.posZ = wheel.radius;
							if additionalWheel.waterBodyFX_isCrawler then water_psData.posZ = crawlers[1].trackWidth*3 end;
							ParticleUtil.loadParticleSystemFromData(water_psData, additionalWheel.water_psR, nil, false, nil, waterBodyFX.dir, wheel.waterBodyFXMainNode);
							local rX,rY,rZ = getRotation(additionalWheel.water_psR.shape);
							setRotation(additionalWheel.water_psR.shape, rX,rY-math.rad(180),rZ);
							setScale(additionalWheel.water_psF.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							setScale(additionalWheel.water_psR.shape, additionalWheel.width * scaleFactor, additionalWheel.radius * scaleFactor, additionalWheel.radius * scaleFactor);
							additionalWheel.water_psF.resetPos = {getTranslation(additionalWheel.water_psF.shape)}
							additionalWheel.water_psR.resetPos = {getTranslation(additionalWheel.water_psR.shape)}
						end;
					end;
				end;
			end;
		end;
		if waterBodyFX.printVehiclesToLog and self.waterBody_FX ~= nil then print("Load mod: '"..waterBodyFX.title.."' successfully applied to "..FSBaseMission:getVehicleName(self)); end;
	end;
end;

function waterBodyFX:getIsCrawler(self, wheel, crawlers)
	if wheel.wheelTire == nil then
		if self.waterBody_FX.crawlers ~= nil then
			if wheel.tireType == WheelsUtil.getTireType("crawler") then 
				return true;
			else
				for _,crawler in pairs(crawlers) do
					if crawler.wheel == wheel then 
						return true; 
					end;
				end;
			end;
		end;
	end;
	return false;
end;

function waterBodyFX:getHasIssue(self, index, wheel)
	if not wheel.hasParticles and wheel.wheelTire == nil then 
		return true, 0; 
	end;
	---------------
	if self.spec_plow ~= nil or self.spec_cultivator ~= nil or self.spec_roller then 
		return true, 0; 
	end;
	---------------
	local namesToFind = {"jcb fastrac 4220"}; --wheel[4] has local offset from rotation
	local vName = FSBaseMission:getVehicleName(self);
	vName = string.lower(vName);
	for _, nameToFind in pairs(namesToFind) do
		if string.find(vName,nameToFind) and index > 3 then 
			return false, 1;
		end;
	end;
	---------------
	local storeCat = g_storeManager:getItemByXMLFilename(self.configFileName);
	if storeCat ~= nil and storeCat.categoryName ~= nil then
		storeCat = storeCat.categoryName;
		for _, type in pairs({ "PLANTERS", "SEEDERS", "WEEDERS", "TEDDERS",
		"WINDROWERS", "PLOWS", "SUBSOILERS", "CULTIVATORS", "DISCHARROWS", "POWERHARROWS" }) do
			if type == storeCat then 
				return true, 0;
			end;
		end;
	end;
	---------------
	return false, 0;
end;

function waterBodyFX:getIsTireInWater(wheel)
	if g_currentMission.waterBodyFX ~= nil and g_currentMission.waterBodyFX.waterY ~= nil then
		local _,wheelY,_ = getWorldTranslation(wheel.waterBodyFXMainNode);
		if wheel.mudFXisActive == nil or not wheel.mudFXisActive then -- Real Mud mod
			if wheelY <= g_currentMission.waterBodyFX.waterY then
				return true;
			end;
		end;
	end;
	return false;
end;

function waterBodyFX:stopPS(wheel)
	if wheel.water_psF ~= nil then ParticleUtil.setEmittingState(wheel.water_psF, false, false, false); end;
	if wheel.water_psR ~= nil then ParticleUtil.setEmittingState(wheel.water_psR, false, false, false); end;
	wheel.waterBodyFXisActive = false;
end;

function waterBodyFX:stopEveryPS(self)
	if self.waterBody_FX ~= nil then
		local wheels = self.spec_wheels.wheels;
		if wheels ~= nil and #wheels > 0 and not self.waterBody_FX.tooManyWheelsForwater then
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.waterBodyFXisActive ~= nil then
					waterBodyFX:stopPS(wheel);
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							waterBodyFX:stopPS(additionalWheel);
						end;
					end;
				end;
			end;
		end;
	end;
end;

function waterBodyFX:correctPSLocation(wheel)
	local _,y,_ = getTranslation(wheel.water_psF.shape);
	local _,y2,_ = getTranslation(wheel.water_psR.shape);
	setTranslation(wheel.water_psF.shape, wheel.water_psF.resetPos[1], y, wheel.water_psF.resetPos[3]);
	setTranslation(wheel.water_psR.shape, wheel.water_psR.resetPos[1], y2, wheel.water_psR.resetPos[3]);
end;

function waterBodyFX:setPSHeight(wheel)
	local x,y,z = getWorldTranslation(wheel.waterBodyFXMainNode);
	local gY = g_currentMission.waterBodyFX.waterY;
	if gY > (y+(wheel.radius*2)) then
		gY = DensityMapHeightUtil.getHeightAtWorldPos(x,y,z);
	end;
	local xT,_,zT = getWorldTranslation(wheel.water_psF.shape);
	local xT2,_,zT2 = getWorldTranslation(wheel.water_psR.shape);
	setWorldTranslation(wheel.water_psF.shape, xT,gY,zT);
	setWorldTranslation(wheel.water_psR.shape, xT2,gY,zT2);
	waterBodyFX:correctPSLocation(wheel);
	if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
		for _, aWheel in pairs(wheel.additionalWheels) do
			local xx,_,zz = getWorldTranslation(aWheel.water_psF.shape);
			local xx2,_,zz2 = getWorldTranslation(aWheel.water_psR.shape);
			local xx3,yy3,zz3 = localToWorld(aWheel.water_psF.shape,0,0,aWheel.radius/2)
			gY = DensityMapHeightUtil.getHeightAtWorldPos(xx3,yy3,zz3);
			setWorldTranslation(aWheel.water_psF.shape, xx,gY,zz);
			setWorldTranslation(aWheel.water_psR.shape, xx2,gY,zz2);
			waterBodyFX:correctPSLocation(aWheel);
		end;
	end;
end;

function waterBodyFX:updateWheelRotation(self, wheel)
	if self.waterBody_FX.crawlers == nil or wheel.wheelTire ~= nil then
		setRotation(wheel.waterBodyFXMainNode,0,wheel.steeringAngle,0);
	end;
end;

function waterBodyFX:updatePS(self, dt)
	local speedfactor = waterBodyFX.psSpeedRatio;
	local enteredSpec = self.spec_enterable;
	local attachableSpec = self.spec_attachable;
	local isEntered = false;
	local isAttachment = false;
	if enteredSpec ~= nil then isEntered = Utils.getNoNil(enteredSpec.isEntered, false); end;
	if not isEntered then
		if attachableSpec ~= nil and attachableSpec.attacherVehicle ~= nil and attachableSpec.attacherVehicle.spec_enterable then
			isEntered = Utils.getNoNil(attachableSpec.attacherVehicle.spec_enterable.isEntered, false);
			isAttachment = true;
		end;
	end;
	if isEntered and not self.tooManyWheelsForwater then
			local wheels = self.spec_wheels.wheels;
			for w=1, #wheels do
				local wheel = wheels[w];
				if wheel.waterBodyFXisActive ~= nil then
					local tspeed = 0;
					if isAttachment then tspeed = self:getLastSpeed(true);
					elseif wheel.netInfo.xDriveSpeed ~= nil then 
						tspeed = MathUtil.rpmToMps(wheel.netInfo.xDriveSpeed/(2*math.pi)*60,wheel.radius)*3600;
						if tspeed < 0 then tspeed = tspeed*-1; end;
					end;
					if not wheel.waterBodyFXisActive then   --[[Activate]]
						if tspeed > 4.0 and waterBodyFX:getIsTireInWater(wheel) then
							if wheel.contact ~= Wheels.WHEEL_NO_CONTACT then wheel.waterBodyFXisActive = true; end;
						end;
					else
						waterBodyFX:setPSHeight(wheel);
						----------------------------
						if tspeed <= 4.0 then
							waterBodyFX:stopPS(wheel);
						----------------------------
						elseif self.movingDirection < 0 then
						----------------------------Update Reverse--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.water_psF, false);
							ParticleUtil.setEmittingState(wheel.water_psR, true);
							ParticleUtil.setParticleSystemSpeed(wheel.water_psR, tspeed/(speedfactor/(wheel.radius*2)));
						elseif self.movingDirection > 0 then
						----------------------------Update Forwards--------------------------------------------------
							ParticleUtil.setEmittingState(wheel.water_psR, false);
							ParticleUtil.setEmittingState(wheel.water_psF, true);
							ParticleUtil.setParticleSystemSpeed(wheel.water_psF, tspeed/(speedfactor/(wheel.radius*2)));
						end;
					end;
					if not waterBodyFX:getIsTireInWater(wheel) then
						waterBodyFX:stopPS(wheel);
						waterBodyFX:setPSHeight(wheel);
					end;
					------------------------------------------------------------------
					if wheel.additionalWheels ~= nil and #wheel.additionalWheels > 0 then
						for _,additionalWheel in pairs (wheel.additionalWheels) do
							if not additionalWheel.waterBodyFXisActive then
								if tspeed > 4.0 and waterBodyFX:getIsTireInWater(additionalWheel) then
									additionalWheel.waterBodyFXisActive = true;
								end;
							else
								----------------------------
								if tspeed <= 4.0 then
									waterBodyFX:stopPS(additionalWheel);
								----------------------------
								elseif self.movingDirection < 0 then
								----------------------------Update Reverse--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.water_psF, false);
									ParticleUtil.setEmittingState(additionalWheel.water_psR, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.water_psR, tspeed/(speedfactor/(additionalWheel.radius*2)));	
								elseif self.movingDirection > 0 then
								----------------------------Update Forwards--------------------------------------------------
									ParticleUtil.setEmittingState(additionalWheel.water_psR, false);
									ParticleUtil.setEmittingState(additionalWheel.water_psF, true);
									ParticleUtil.setParticleSystemSpeed(additionalWheel.water_psF, tspeed/(speedfactor/(additionalWheel.radius*2)));
								end;
							end;
							if not waterBodyFX:getIsTireInWater(additionalWheel) then 
								waterBodyFX:stopPS(additionalWheel);
							end;
						end;
					end;
					----------------------------------------------------------------------
					waterBodyFX:updateWheelRotation(self, wheel);
				end;
			end;
	elseif not self.tooManyWheelsForwater then
		waterBodyFX:stopEveryPS(self);
	end;

end;

function waterBodyFX:updateAccelLimit(self)
	local _,y,_ = getWorldTranslation(self.rootNode); y = y+1;
	if y < g_currentMission.waterBodyFX.waterY then
		local multiplier = self.waterBody_FX.defaultAccelLimit / 2;
		y = ((g_currentMission.waterBodyFX.waterY-y)/2.5)*multiplier;
		self:getMotor():setAccelerationLimit(math.max(self.waterBody_FX.defaultAccelLimit-y,0.25));
	elseif self:getMotor():getAccelerationLimit() ~= self.waterBody_FX.defaultAccelLimit then
		self:getMotor():setAccelerationLimit(self.waterBody_FX.defaultAccelLimit);
	end;
end;

function waterBodyFX:onUpdate(dt)
	if self.spec_wheels ~= nil then
		if self.spec_wheels.wheels ~= nil and #self.spec_wheels.wheels > 0 then
			waterBodyFX:updatePS(self, dt);
		end;
	end;
	if self.spec_motorized ~= nil then
		waterBodyFX:updateAccelLimit(self);
	end;
end;

function waterBodyFX:onLeave()
	waterBodyFX:stopEveryPS(self);
end;

function waterBodyFX:disable()
	waterBodyFX.disabled=1;
end;

if waterBodyFX.disabled == nil then
	print("Load mod: "..waterBodyFX.title.." : v"..waterBodyFX.version.." by "..waterBodyFX.author);
end;